/*
 +========================================================================+
 |         (c) 2020 - PROPERTY OF ECOLE POLYTECHNIQUE - LGPL 3.0          |
 |________________________________________________________________________|
 |   '&`   |                                                              |
 |    #    |   FILE       : dirichletBW.cpp                               |
 |    #    |   VERSION    : 0.1.0                                         |
 |   _#_   |   AUTHOR(S)  : Matthieu Aussal                               |
 |  ( # )  |   CREATION   : 01.04.2020                                    |
 |  / 0 \  |   LAST MODIF : 31.10.2020                                    |
 | ( === ) |   SYNOPSIS   : Acoustical scattering using BEM with Dirichlet|
 |  `---'  |                boundary condition                            |
 +========================================================================+
 */

#include <castor/matrix.hpp>
#include <castor/smatrix.hpp>
#include <castor/hmatrix.hpp>
#include <castor/linalg.hpp>
#include <castor/graphics.hpp>
#include "bembuilder.hpp"

using namespace castor;

int main (int argc, char* argv[])
{
    // Load meshes
    matrix<double> Svtx, Pvtx;
    matrix<size_t> Stri, Ptri;
    std::tie(Stri,Svtx) = triread("/Users/aussal/Git/leprojetcastor/bembuilder/data/","unitSphere_1e3.ply");
    std::tie(Ptri,Pvtx) = triread("/Users/aussal/Git/leprojetcastor/bembuilder/data/","Oxyz_1e4.ply");
    
    // Graphical representation
    figure fig;
    trimesh(fig,Ptri,Pvtx);
    trimesh(fig,Stri,Svtx);
    
    // Parameters
    matrix<double> U = {-1,0,0};
    double k = 5.85;
    float tol = 1e-3;
    
    // FEM
    tic();
    femdata<double> v(Stri,Svtx,lagrangeP1,3);
    femdata<double> u(Stri,Svtx,lagrangeP1,3);
    toc();

    // Incident wave : PW(x,u) = exp(i*k*x.u)
    auto Pinc = planeWave(Pvtx,U,k);
    
    // Identity matrix
    tic();
    auto Id = mass<std::complex<double>>(v);
    toc();

    // Coupling coeff
    std::complex<double> beta(0,0.5*k);
    
    // Left hand side : [1i*k*beta*S - (Id/2 + D)]
    tic();
    auto LHSfct = [&v,&u,&k,&beta,&Id](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return beta*singleLayer<std::complex<double>>(v,u,k,Ix,Iy)
        - (0.5*eval(Id(Ix,Iy)) + doubleLayer<std::complex<double>>(v,u,k,Ix,Iy));
    };
    hmatrix<std::complex<double>> LHS(v.dof(),u.dof(),tol,LHSfct);
    toc();
    disp(LHS);

    // Right hand side : - \int_sigma_x v(x) PW(x,u) dx
    auto B = - rightHandSide<std::complex<double>>(v,PWsource,U,k);

    // Solve linear system : [1i*k*beta*S - (Id/2 + D)] = P0
    auto mu     = gmres(LHS,B,tol,100);
    auto lambda = beta * mu;
    
    // Boundary radiation
    tic();
    auto Sbndfct = [&v,&u,&k,&Id](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return singleLayer<std::complex<double>>(v,u,k,Ix,Iy);
    };
    auto Dbndfct = [&v,&u,&k,&Id](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return 0.5*eval(Id(Ix,Iy)) + doubleLayer<std::complex<double>>(v,u,k,Ix,Iy);
    };
    matrix<std::complex<double>> Pbnd = zeros(size(v.dof(),1),size(mu,2));
    hmatrix<std::complex<double>> Sbnd(v.dof(),u.dof(),tol,Sbndfct,lambda,Pbnd);
    hmatrix<std::complex<double>> Dbnd(v.dof(),u.dof(),tol,Dbndfct,-mu,Pbnd);
    toc();
    Pbnd = gmres(Id,Pbnd,tol,100) + planeWave(v.dof(),U,k);
    
    // Domain radiation
    tic();
    auto Sdomfct = [&Pvtx,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return singleLayer<std::complex<double>>(Pvtx,u,k,Ix,Iy);
    };
    auto Ddomfct = [&Pvtx,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return doubleLayer<std::complex<double>>(Pvtx,u,k,Ix,Iy);
    };
    matrix<std::complex<double>> Pdom = Pinc;
    hmatrix<std::complex<double>> Sdom(Pvtx,u.dof(),tol,Sdomfct,lambda,Pdom);
    hmatrix<std::complex<double>> Ddom(Pvtx,u.dof(),tol,Ddomfct,-mu,Pdom);
    toc();
    for (std::size_t l=0; l<size(Pvtx,1); ++l)
    {
        if (std::pow(Pvtx(l,0),2)+std::pow(Pvtx(l,1),2)+std::pow(Pvtx(l,2),2)<1.1) {Pdom(l)=Pinc(l);}
    }
    
    // Graphical representation
    figure fig2;
    trimesh(fig2,Ptri,Pvtx,abs(Pdom));
    trimesh(fig2,Stri,Svtx,abs(Pbnd));

    // Infinite radiation
    tic();
    matrix<double> theta = transpose(linspace(0,2*M_PI,1e3));
    matrix<double> Nu = horzcat(horzcat(cos(theta),sin(theta)),zeros(size(theta)));
    auto Sinffct = [&Nu,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return singleLayerInfinite<std::complex<double>>(Nu,u,k,Ix,Iy);
    };
    auto Dinffct = [&Nu,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return doubleLayerInfinite<std::complex<double>>(Nu,u,k,Ix,Iy);
    };
    matrix<std::complex<double>> Pinf = zeros(size(Nu,1),size(lambda,2));
    hmatrix<std::complex<double>> Sinf(Nu,u.dof(),0,Sinffct,lambda,Pinf);
    hmatrix<std::complex<double>> Dinf(Nu,u.dof(),0,Dinffct,-mu,Pinf);
    toc();
    
    // Graphical representation
    figure fig3;
    plot(fig3,theta,transpose(log(abs(Pinf))),{"-b"});
    
    // Plot
    drawnow(fig);
    disp("done !");
    return 0;
}
